import java.util.*;

/**
 * propositionele formules 
 */
abstract public class AbstractFormule implements Formule {
  
  /**
   * connectieven in oplopende ariteit en (niet noodzakelijk strikt) dalende bindingssterkte
   */
  enum Connectief { ATOOM,ONWAAR,WAAR,NIET,EXOF,EN,OF,ALSDAN,DESDA }
  
  /**
   * we representeren een formule als een aantal subformules  (inductie) en een connectief
   * als het een atoom is, dan heeft de formule ook een index
   */
     
  private AbstractFormule[] subFormules;
  private Connectief connectief;
  private Integer index;
  
  public AbstractFormule(Connectief connectief,AbstractFormule ... subFormules) { 
    this.connectief = connectief; 
    this.subFormules = subFormules; }
  
  public AbstractFormule(Integer index) {
    this.connectief = Connectief.ATOOM;
    this.index = index;
  }
  
  public AbstractFormule[] getSubFormules(){ return subFormules; }
  
  public Connectief getConnectief(){ return connectief; }
  
  public int getIndex() { return index; } 
  
  /**
   * @return formule als String, de haakjes conventies respecterend
   */
  public String toString() {
    String string = (connectief == Connectief.ATOOM) ? "P"+index : connectief.toString();
    return (connectief.ordinal() < Connectief.NIET.ordinal()) ?
    string : ((connectief == Connectief.NIET)  ? string+" "+toString(0) : toString(0)+" "+string+" "+toString(1));
  }
  
  /**
   * @return i-de subformule als String, evt. voorzien van haakjes mochten de haakjesconventies dat vereisen
   */
  public String toString(int i) {
    String string = subFormules[i].toString();
    return ((subFormules[i].getConnectief().ordinal() <= Connectief.NIET.ordinal()) 
            | ((subFormules[i].getConnectief().ordinal() <= Connectief.OF.ordinal()) & (connectief.ordinal() > Connectief.OF.ordinal())))
    ? string : "("+string+")";
  }
  
  /**
   * @return bedeling wordt in String gezet (alleen voor debug doeleinden)
   */
  public static String toString(int[] bedeling) {
    String string = "[";
    for (int teller = 0;teller < bedeling.length;teller++) 
      string += ((teller==0)?"":",")+"P"+((bedeling[teller]<0)?-bedeling[teller]+" := onwaar":bedeling[teller]+":= waar");
    return string+"]";
  }
  
  /**
   * @return de Herbrand normaalvorm van deze formule als Herbrand formule
   */
  abstract public HerbrandFormule toHerbrandFormule(); // geschikt te overriden in de subklassen per connectief

  /**
   * @return de Herbrand normaalvorm van deze formule
   */
  public Formule herbrandNormaalvorm() { 
    return toHerbrandFormule().toFormule(); // omzetten naar Herbrand formule (dat doet het werk) en terug 
  }
  
  /**
   * @return update van de atomen van deze formule met de bedeling
   */
  public void bedelen(int[] bedeling) {
    if (connectief == Connectief.ATOOM) {
      for (int index:bedeling) 
        if ((this.index.intValue() == index) || (this.index.intValue() == -index)) 
          this.connectief = (this.index.intValue() == index) ? Connectief.WAAR : Connectief.ONWAAR;
    } else {
      for (AbstractFormule formule:getSubFormules())
        formule.bedelen(bedeling);
    }
  }

  /**
   * @return waarheidswaarde van formule onder bedeling
   */
  public boolean waarde(int[] bedeling) throws OnvoldoendeBedeling {
    AbstractFormule formule = toHerbrandFormule().toFormule(); // een kopie van de formule maken, door heen en weer vertalen
    formule.bedelen(bedeling); // atomen veranderen in hun waarheidswaarde, voor zover in bedeling
    HerbrandFormule herbrandFormule = formule.toHerbrandFormule(); // Herbrand normaalvorm bepalen
    if (herbrandFormule.getClauses().size() == 0) return false; // onwaar als een lege verzameling clauses
    if ((herbrandFormule.getClauses().size() == 1) && (herbrandFormule.getClauses().first().getIndices().size() == 0)) return true;
    // waar als een enkele lege clause
    throw new OnvoldoendeBedeling();
  }
  
  /**
   * @return de verzameling van de indices van de atomen in deze formule, geordend op index, zonder herhaling
   */
  public SortedSet<Integer> indices() {
    SortedSet<Integer> indices = new TreeSet<Integer>();
    if (connectief == Connectief.ATOOM) {
      indices.add(index);
    } else {
      for (AbstractFormule formule:getSubFormules())
        indices.addAll(formule.indices());
    }
    return indices;
  }
    
  /**
   * @return een vervulling van de formule zo deze bestaat, anders null.
   */    
  public int[] vervulbaar() { 
    HerbrandFormule herbrandFormule = toHerbrandFormule();// eerst naar Herbrand normaalvorm om overbodige variabelen weg te werken
    AbstractFormule formule = herbrandFormule.toFormule(); 
    SortedSet<Integer> indices = formule.indices(); // verzameling van variabelen in formule
    int[] bedeling = new int[indices.size()];
    int teller = 0;
    for (Integer index:indices) 
      bedeling[teller++] = index; // variabelen omzetten naar een bedeling;
    while (!formule.waarde(bedeling)) {
      teller = 0;
      while (teller < bedeling.length) {
        bedeling[teller] = -bedeling[teller];
        if (bedeling[teller] < 0) 
          break;
        teller++;
      }
      if (teller == bedeling.length) 
        break;
    }
    return (formule.waarde(bedeling)) ? bedeling : null;
  }
}
